/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * ad.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * A/D converter control and processing
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>

#include "ad.h"
#include "def.h"
#include "fault.h"

extern volatile uint8_t  rssi_reading;
extern volatile uint32_t instr_state;




/*
************************************************************************************************************************************************
RSSI A/D Converter:
These routines read out the A/D converter to calculate the RSSI display reading in dBm in when receiver is in AGC mode or in percent 
demodulator usage when receiver is in MGC mode.
*/

static const uint16_t gainctrl_scaledivisor = 4;											// Scale factor for gain control voltage level 
static const uint16_t rssi_scaledivisor = 8;												// Scale factor for RSSI voltage level 
static const int16_t  offset_dbm = 215;														// Offset for signal level calculation from ADC results

static const uint16_t rssi_mgcscaledivisor = 2;												// Scale factor for percent demodulator usage calculation from ADC result
static const int16_t  offset_mgc = 200;														// Offset for percent demodulator usage calculation from ADC result

static const uint8_t att_value = 20;														// Attenuation of switchable attenuator in dB
static const uint8_t preamp_value = 25;														// Preamplifier gain in db
static const uint8_t presel_loss = 1;														// Preselector loss in dB

static volatile uint16_t gainctrl_meas[4];													// The last four measured gain control voltages
static volatile uint16_t rssi_meas[4];														// The last four measured RSSI voltages
static volatile uint8_t  adc_count;															// Counts the ADC measuring results for averaging
static volatile uint8_t  adc_finished;														// Flag to be set when ADC has finished, reset when result is read out

void adc_init(void)
{
	TCCR3B |= (1<<WGM32);																	// Clear Timer3 on compare
	TCCR3B |= (1<<CS31) | (1<<CS30);														// Set Timer3 prescaler to divide by 64
	ETIMSK  |= (1<<OCIE3A);																	// Timer3 output compare match A interrupt enable	
	OCR3AH = 0x0E;
	OCR3AL = 0xA8;																			// Compare register A set to (3752)_dec = (0EA8)_hex, interrupt approx. every 15ms
	
	ADCSRA |= (1<<ADEN) | (1<<ADIE);														// Enable ADC and ADC interrupt
	ADCSRA |= (1<<ADPS0) | (1<<ADPS1) | (1<<ADPS2);											// Select ADC prescaler to divide by 128 (= 62.5 kHz ADC clock)
	ADCSRA &= ~( (1<<ADATE) );																// ADC auto trigger disable
	ADMUX  &= ~( (1<<REFS0) | (1<<REFS1) );													// Use external reference voltage (Vref = 3.3V)
	ADMUX  &= ~( (1<<ADLAR) );																// A/D result right adjust
	ADMUX  &= ~( (1<<MUX0) | (1<<MUX1) | (1<<MUX2) | (1<<MUX3) | (1<<MUX4) );				// Set ADC multiplexer (clear all mux bits)
}

ISR(TIMER3_COMPA_vect)																		// ISR starts AD conversion
{
	if( !adc_finished && (adc_count <= 7) )													// Have ADC results been read out and less than 8 conversions have been accumulated?
		ADCSRA |= (1<<ADSC);																// Start one AD conversion;
}

ISR(ADC_vect)																				// ADC interrupt service routine
{
	adc_finished = 1;																		// Set flag
}

void adc_process(void)																		// Processes ADC result if four conversions have been accumulated, compute average and write to appropriate global
{
	int16_t gainctrl_average;
	int16_t rssi_average;
	int16_t val;
	
	if( adc_finished && (adc_count < 8) )													// ADC has finished and less than 8 conversions accumulated?
	{
		if( (adc_count & 1) == 0 )															// Is count is even?
		{
			rssi_meas[adc_count/2] = ADC;													// Store RSSI voltage result
			ADMUX |= (1<<MUX0);																// Select ADC input 1 for gain control voltage measurement			
		}
		else																				// count is odd
		{
			gainctrl_meas[adc_count/2] = ADC;												// Store gain control voltage result
			ADMUX &= ~( (1<<MUX0) );														// Select ADC input 0 for RSSI voltage measurement			
		}
		adc_count++;
		adc_finished = 0;																	// Reset ADC finished flag to indicate that result has been stored	
	}
	
	if( adc_count >= 8 )																	// After eight ADC conversions have been stored process results
	{
		gainctrl_average = ( gainctrl_meas[0] + gainctrl_meas[1] + gainctrl_meas[2] + gainctrl_meas[3] ) / 4;
		rssi_average = ( rssi_meas[0] + rssi_meas[1] + rssi_meas[2] + rssi_meas[3] ) / 4;	// Calculate average of four results
		adc_count = 0;																		// Reset number of A/D conversions counter
		
		if( !(instr_state & (1UL<<MGC)) )													// Is receiver in AGC mode? Then calculate RSSI display reading in dBm from rssi_meas and gainctrl_meas
		{
			val = offset_dbm - ( (gainctrl_average/gainctrl_scaledivisor) + (rssi_average/rssi_scaledivisor) );
			if( instr_state & (1UL<<ATT_IN) )												// Is switchable attenuator in?
				val -= att_value;
			if( instr_state & (1UL<<PREAMP_IN) )											// Is preamplifier in?
				val += preamp_value;
			if( instr_state & (1UL<<PRESEL_IN) )											// Is preselector in?
				val -= presel_loss;
			if( val > 0 )																	// Is calculated RSSI value to be displayed larger than 0?
				rssi_reading = (uint8_t)( val );
			else
				rssi_reading = 0;
		}
		else																				// Receiver is in MGC mode, calculate RSSI display reading as percent (0 to 100) demodulator usage from rssi_meas only
		{
			val = (rssi_average/rssi_mgcscaledivisor) - offset_mgc;
			if( (val > 0) && (val < 100) )
				rssi_reading = (uint8_t)( val );
			else if( val >= 100 )
				rssi_reading = 100;
			else
				rssi_reading = 0;
		}
		
		if( !( rssi_average < 120 ) )														// Check if RSSI voltage is in permissible range (i.e. greater than approx 0.4V), else flag an error
			clear_error(ERR_RSSI_RANGE);
		else
			handle_error(ERR_RSSI_RANGE);
	}
}
