/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * annunciators.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * Annunciator multiplexing and output
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>

#include "annunciators.h"
#include "def.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  annunc_pattern[16];
extern volatile uint32_t tuned_freq;
extern volatile int16_t  bfo_offset;
extern volatile uint8_t  cor_level;
extern volatile uint8_t  rssi_reading;
extern volatile struct
{
	uint8_t digit[8];
	uint8_t decpoint_pos;
	uint8_t decpoint_flag;
	uint8_t	digit_count;
	uint8_t sign;
} scratchpad;




/*
************************************************************************************************************************************************
Display and annunciator multiplexing:
Multiplexing is performed digit-wise in the seven segment displays and is triggered by Timer0. The annunciators in pushbuttons numerical 
displays are combined in four groups of eight LEDs (the last group is partially empty), the groups are illuminated consecutively after the
seven segment displays. Multiplexing clock is 1ms.
*/

static volatile uint8_t mux_count = 0;

#define ANNUNC_PORT				PORTA
#define ANNUNC_DDR				DDRA
#define MUX_PORT				PORTB
#define MUX_DDR					DDRB
#define MUX_OUT0				PORTB4
#define MUX_OUT1				PORTB5
#define MUX_OUT2				PORTB6
#define MUX_OUT3				PORTB7
#define ANNUNC_DEACTIVATE		PORTA = 0x00


void annunc_init(void)
{
	ANNUNC_DDR = 0xFF;																		// Set annunciator pins as output
	MUX_DDR   |= (1<<MUX_OUT0) | (1<<MUX_OUT1) | (1<<MUX_OUT2) | (1<<MUX_OUT3);				// Set multiplex pins as output
	ANNUNC_DEACTIVATE;																		// CLear annunciator output
	MUX_PORT &= ~( (1<<MUX_OUT0) | (1<<MUX_OUT1) | (1<<MUX_OUT2) | (1<<MUX_OUT3) );			// Clear multiplex output
	TCCR0 |= (1<<CS00) | (1<<CS02);															// Timer0 prescaler set to divide by 128
	TCCR0 |= (1<<WGM01);																	// Clear timer on compare
	TIMSK |= (1<<OCIE0);																	// Timer0 output compare match interrupt enable
	OCR0 = 125;																				// Compare Register, set to 125, interrupt approx every 1ms
}

ISR(TIMER0_COMP_vect)																		// Interrupt service routine for Timer0 compare match
{
	ANNUNC_DEACTIVATE;																		// Turn off annunciators before changing multiplex output to help avoid ghosting
	NOP();																					// Wait one clock cycle to help avoid ghosting
	MUX_PORT = (MUX_PORT & 0x0F) | (mux_count << 4);										// Update multiplex output
	NOP();
	ANNUNC_PORT = annunc_pattern[mux_count];												// Update annunciator output
	if(mux_count < 15)																		// mux_count is less than maximal value?
		++mux_count;																		// Then increment
	else
		mux_count = 0;																		// Else reset to 0
}




/*
************************************************************************************************************************************************
Data Output to Seven Segment Displays and Annunciators
*/

// Seven segment display patterns
#define DISP_0					0b00111111    // 0: a, b, c, d, e, f
#define DISP_1					0b00000110    // 1: b, c
#define DISP_2					0b01011011    // 2: a, b, d, e, g
#define DISP_3					0b01001111    // 3: a, b, c, d, g
#define DISP_4					0b01100110    // 4: b, c, f, g
#define DISP_5					0b01101101    // 5: a, c, d, f, g
#define DISP_6					0b01111101    // 6: a, c, d, e, f, g
#define DISP_7					0b00000111    // 7: a, b, c
#define DISP_8					0b01111111    // 8: a, b, c, d, e, f, g
#define DISP_9					0b01101111    // 9: a, b, c, d, f, g
#define DISP_BLANK				0b00000000	  // Blank
#define DISP_DP					0b10000000	  // Decimal Point
#define DISP_MINUS				0b01000000	  // Minus or Bar

static uint8_t segments (uint8_t digit)
{
	switch(digit)
	{
		case 0:
			return DISP_0;
			break;
		case 1:
			return DISP_1;
			break;
		case 2:
			return DISP_2;
			break;
		case 3:
			return DISP_3;
			break;
		case 4:
			return DISP_4;
			break;
		case 5:
			return DISP_5;
			break;
		case 6:
			return DISP_6;
			break;
		case 7:
			return DISP_7;
			break;
		case 8:
			return DISP_8;
			break;
		case 9:
			return DISP_9;
			break;
		default:
			return DISP_BLANK;
			break;
	}
}

static void display_freq(uint32_t freq)														// Writes freq to annunc_pattern for display on main frequency display (eight digits, in MHz)
{
	uint8_t i;
	for ( i = 8; i != 0; --i )																// Update frequency display by writing to annunc_pattern, loop over digits
	{
		if( !( (i-1 == 0) && (freq == 0) ) )												// Suppress leading zeros: blank digit if freq is zero on MSD
		{
			cli();
			annunc_pattern[i-1] = segments( freq % 10 );									// Write digit to annunc_pattern, global interrupts off until digit is written
			if( i-1 == 1 )																	// Is the current digit the second from left (MSD-1)?
				annunc_pattern[i-1] |= DISP_DP;												// Then set decimal point
			sei();
		}
		else
			annunc_pattern[i-1] = DISP_BLANK;
		freq = freq / 10;																	// Divide by 10 to extract next digit from freq
	}
}

static void display_bfo_offs(int16_t bfo_offs)												// Writes bfo_offs to annunc_pattern for display on main frequency display (four digits, in kHz)
{
	uint8_t i;
	uint8_t sign;
	if( bfo_offs < 0 )																		// BFO offset is negative?
	{
		bfo_offs = -bfo_offs;																// Then change sign
		sign = 1;																			// Remember sign
	}
	else
		sign = 0;
	for ( i = 7; i >= 4; --i )																// Update BFO display, BFO offset is displayed on the four LSD
	{
		cli();
		annunc_pattern[i] = segments( bfo_offs % 10 );										// Write digit to annunc_pattern, global interrupts off until digit is written
		if( i == 4 )																		// Is the current digit the fourth from right?
			annunc_pattern[i] |= DISP_DP;													// Then set decimal point
		bfo_offs = bfo_offs / 10;															// Divide by 10 to extract next digit from bfo_offs
		sei();
	}
	annunc_pattern[0] = DISP_BLANK;
	annunc_pattern[1] = DISP_BLANK;
	annunc_pattern[2] = DISP_BLANK;															// Blank the remaining digits on the left
	if( sign == 0 )																			// Is bfo_offs nonnegative?
		annunc_pattern[3] = DISP_BLANK;														// Then display no sign (blank digit)
	else
		annunc_pattern[3] = DISP_MINUS;														// Else display a minus
}

static void display_rssi_reading(uint8_t rssi_reading)										// Writes rssi_reading to annunc_pattern for display on RSSI display
{
	uint8_t i;
	for ( i = 3; i != 0; --i )																// Update RSSI display
	{
		if( !( (i+8 <= 10) && (rssi_reading == 0) ) )										// Suppress leading zeros: blank digit if rssi_reading is zero on the two MSD
			annunc_pattern[i+8] = segments( rssi_reading % 10 );							// Write digit to annunc_pattern
		else
			annunc_pattern[i+8] = DISP_BLANK;
		rssi_reading = rssi_reading / 10;													// Divide by 10 to extract next digit from rssi_reading
	}
	if( !(instr_state & (1UL<<MGC)) )														// Receiver not in manual gain mode?
		annunc_pattern[8] = DISP_MINUS;														// Then display sign on MSD digit of RSSI display (RSSI reading is in negative dBm)
	else																					// Receiver is in manual gain mode
		annunc_pattern[8] = DISP_BLANK;														// Blank first digit on RSSI display
}

static void display_cor_level(uint8_t cor_level)											// Writes cor_level to annunc_pattern for display on RSSI display, and checks for limits
{
	uint8_t i;
	if ( cor_level >= MIN_COR_LEVEL )														// Is COR level below minimum?
	{
		annunc_pattern[8]  = DISP_MINUS;													// Then display "----"
		annunc_pattern[9]  = DISP_MINUS;
		annunc_pattern[10] = DISP_MINUS;
		annunc_pattern[11] = DISP_MINUS;
	}
	else																					// COR level is above minimum
	{
		for ( i = 3; i != 0; --i )															// Update COR level display
		{
			annunc_pattern[i+8] = segments( cor_level % 10 );
			cor_level = cor_level / 10;
		}
		annunc_pattern[8] = DISP_MINUS;														// Display sign on MSD digit of RSSI display
	}
}

static void display_scratchpad(void)														// Displays current scratchpad contents for numeric keypad entry when keyboard is shifted
{
	static uint8_t last_digits;
	static uint8_t last_sign;
	int8_t i;
	int8_t j;
	if( scratchpad.digit_count == 0 )														// No digits entered in scratchpad?
	{
		annunc_pattern[0] = DISP_BLANK;														// Display "_______0." on frequency display
		annunc_pattern[1] = DISP_BLANK;
		annunc_pattern[2] = DISP_BLANK;
		annunc_pattern[3] = DISP_BLANK;
		annunc_pattern[4] = DISP_BLANK;
		annunc_pattern[5] = DISP_BLANK;
		annunc_pattern[7] = DISP_0 | DISP_DP;
		if( !scratchpad.sign )																// Scratchpad value not negative?
			annunc_pattern[6] = DISP_BLANK;													// Display no sign
		else																				// Scratchpad value is negative
			annunc_pattern[6] = DISP_MINUS;													// Display sign
		last_digits = 0;																	// Clear last_digits
	}
	else if( (scratchpad.digit_count != last_digits) || (scratchpad.sign != last_sign) )	// New digit was entered or change sign pressed since last call
	{
		for( i = 0 ; i <= scratchpad.digit_count; i++ )										// Display digits right-aligned on frequency display
			annunc_pattern[8 - scratchpad.digit_count + i] = segments( scratchpad.digit[i] );
		for( j = 0; j < (8 - scratchpad.digit_count); j++ )									// Blank remaining digits
			annunc_pattern[j] = DISP_BLANK;
				
		annunc_pattern[7 - scratchpad.digit_count + scratchpad.decpoint_pos ] |= DISP_DP;	// Dispay decimal point
		
		if( scratchpad.sign )																// Scratch pad value negative?
			annunc_pattern[7 - scratchpad.digit_count] = DISP_MINUS;						// Then display minus
		last_digits = scratchpad.digit_count;												// Update last_digits
		last_sign = scratchpad.sign;														// Update last_sign
	}
}

static void display_error(uint8_t error_disp)												// Displays "Error" and error code on main frequency display
{
	annunc_pattern[0] = 0b01111001;
	annunc_pattern[1] = 0b01010000;
	annunc_pattern[2] = 0b01010000;
	annunc_pattern[3] = 0b01011100;
	annunc_pattern[4] = 0b01010000;															// Display "Error" left aligned on frequency display

	annunc_pattern[8]  = 0x00;
	annunc_pattern[9]  = 0x00;
	annunc_pattern[10] = 0x00;
	annunc_pattern[11] = 0x00;																// Blank RSSI display	
	
	uint8_t i;
	for( i = 7; i >= 5; i-- )																// Display error code on frequency display
	{
		annunc_pattern[i] = segments(error_disp % 10);
		error_disp = error_disp / 10;
	}
}

void disp_out(void)																			// Write output to the seven segment registers, to be called periodically in main function
{
	if( !instr_error )																		// No error flag is set?
	{
		// Main frequency display:
		if( !(instr_state & (1UL<<BFO_TUNE)) && !(instr_state & (1UL<<SHIFT)) )				// Receiver not in BFO tune mode and keyboard not shifted?
			display_freq(tuned_freq);														// Then display tuned frequency				
		else if(  (instr_state & (1UL<<BFO_TUNE)) && !(instr_state & (1UL<<SHIFT)) )		// Receiver in BFO tune mode and keyboard is not shifted?
			display_bfo_offs(bfo_offset);													// Then display BFO offset
		else if( instr_state & (1UL<<SHIFT) )												// Keyboard shifted?
			display_scratchpad();															// Then display scratchpad contents
		// RSSI display:
		if( !(instr_state & (1UL << SET_CORLVL)) )											// Receiver is not in SET COR level mode?
			display_rssi_reading(rssi_reading);												// Then display RSSI reading
		else																				// Receiver is in SET COR level mode
			display_cor_level(cor_level);													// Then display COR level
	}
	else																					// Error flag is set
		display_error(instr_error);															// Display instr_error as decimal value and blank RSSI display	
}

void annunc_out(void)																		// Write annunciator registers, to be called periodically in main function
{
	static uint32_t instr_state_old;
	
	if( instr_state != instr_state_old )													// instr_state has changed?
	{
		instr_state_old = instr_state;
		
		if( !(instr_state & (1UL<<SHIFT)) )													// Keyboard not shifted?
		{
			if( !( instr_state & (1UL<<BFO_TUNE) ) )										// Receiver not in BFO tune mode?
			{
				annunc_pattern[12] = (uint8_t)
				(
				  ( ( (instr_state >> AM)             & 0x01 ) << 0 )
				| ( ( (instr_state >> USB)            & 0x01 ) << 1 )
				| ( ( (instr_state >> LSB)            & 0x01 ) << 2 )
				| ( ( (instr_state >> FM)             & 0x01 ) << 3 )
				| ( ( (instr_state >> AGC_FAST)       & 0x01 ) << 4 )
				| ( ( (instr_state >> AGC_SLOW)       & 0x01 ) << 5 )
				| ( ( (instr_state >> AGC_HANG)       & 0x01 ) << 6 )
				| ( ( (instr_state >> MGC)            & 0x01 ) << 7 )
				);
				annunc_pattern[13] = (uint8_t)
				(
		   		  ( ( (instr_state >> BW1)            & 0x01 ) << 0 )
				| ( ( (instr_state >> BW2)            & 0x01 ) << 1 )
				| ( ( (instr_state >> BW3)            & 0x01 ) << 2 )
				| ( ( (instr_state >> BW4)            & 0x01 ) << 3 )
				| ( ( (instr_state >> TUNE_1HZ)       & 0x01 ) << 4 )
				| ( ( (instr_state >> TUNE_10HZ)      & 0x01 ) << 5 )
				| ( ( (instr_state >> TUNE_100HZ)     & 0x01 ) << 6 )
				| ( ( (instr_state >> TUNE_1KHZ)      & 0x01 ) << 7 )
				);
				annunc_pattern[14] = (uint8_t)
				(
				  ( ( (instr_state >> TUNE_10KHZ)     & 0x01 ) << 0 )
				| ( ( (instr_state >> TUNE_100KHZ)    & 0x01 ) << 1 )
				| ( ( (instr_state >> SET_CORLVL)     & 0x01 ) << 2 )
				| ( ( (instr_state >> BFO_TUNE)       & 0x01 ) << 3 )
				| ( ( (instr_state >> ATT_IN)         & 0x01 ) << 4 )
				| ( ( (instr_state >> PRESEL_IN)      & 0x01 ) << 5 )
				| ( ( (instr_state >> PREAMP_IN)      & 0x01 ) << 6 )
				| ( ( (instr_state >> SPEAKER_ON)     & 0x01 ) << 7 )
				);
				annunc_pattern[15] = (uint8_t)
				(
				  ( ( (instr_state >> SHIFT)          & 0x01 ) << 0 )
				| ( ( (instr_state >> KHZ_LED)        & 0x01 ) << 1 )
				| ( ( (instr_state >> MHZ_LED)        & 0x01 ) << 2 )
				| ( ( (instr_state >> COR_ACTIVE)     & 0x01 ) << 3 )
				);			
			}
			else																			// Receiver in BFO tune mode
			{
				annunc_pattern[12] = (uint8_t)
				(
				  ( ( (instr_state >> AM)             & 0x01 ) << 0 )
				| ( ( (instr_state >> USB)            & 0x01 ) << 1 )
				| ( ( (instr_state >> LSB)            & 0x01 ) << 2 )
				| ( ( (instr_state >> FM)             & 0x01 ) << 3 )
				| ( ( (instr_state >> AGC_FAST)       & 0x01 ) << 4 )
				| ( ( (instr_state >> AGC_SLOW)       & 0x01 ) << 5 )
				| ( ( (instr_state >> AGC_HANG)       & 0x01 ) << 6 )
				| ( ( (instr_state >> MGC)            & 0x01 ) << 7 )
				);
				annunc_pattern[13] = (uint8_t)
				(
				  ( ( (instr_state >> BW1)            & 0x01 ) << 0 )
				| ( ( (instr_state >> BW2)            & 0x01 ) << 1 )
				| ( ( (instr_state >> BW3)            & 0x01 ) << 2 )
				| ( ( (instr_state >> BW4)            & 0x01 ) << 3 )
				| ( ( (instr_state >> BFOTUNE_1HZ)    & 0x01 ) << 4 )
				| ( ( (instr_state >> BFOTUNE_10HZ)   & 0x01 ) << 5 )
				| ( ( (instr_state >> BFOTUNE_100HZ)  & 0x01 ) << 6 )
				);
				annunc_pattern[14] = (uint8_t)
				(
				  ( ( (instr_state >> SET_CORLVL)     & 0x01 ) << 2 )
				| ( ( (instr_state >> BFO_TUNE)       & 0x01 ) << 3 )
				| ( ( (instr_state >> ATT_IN)         & 0x01 ) << 4 )
				| ( ( (instr_state >> PRESEL_IN)      & 0x01 ) << 5 )
				| ( ( (instr_state >> PREAMP_IN)      & 0x01 ) << 6 )
				| ( ( (instr_state >> SPEAKER_ON)     & 0x01 ) << 7 )
				);
				annunc_pattern[15] = (uint8_t)
				(
				  ( ( (instr_state >> SHIFT)          & 0x01 ) << 0 )
				| ( ( (instr_state >> KHZ_LED)        & 0x01 ) << 1 )
				| ( ( (instr_state >> MHZ_LED)        & 0x01 ) << 2 )
				| ( ( (instr_state >> COR_ACTIVE)     & 0x01 ) << 3 )
				);				
			}									
		}
		else																				// Keyboard shifted
		{
			annunc_pattern[12] = 0;
			annunc_pattern[13] = 0;
			annunc_pattern[14] = (uint8_t)
			(
			  ( ( (instr_state >> SET_CORLVL)     & 0x01 ) << 2 )
			| ( ( (instr_state >> BFO_TUNE)       & 0x01 ) << 3 )
			);
			annunc_pattern[15] = (uint8_t)
			(
			  ( ( (instr_state >> SHIFT)          & 0x01 ) << 0 )
			| ( ( (instr_state >> KHZ_LED)        & 0x01 ) << 1 )
			| ( ( (instr_state >> MHZ_LED)        & 0x01 ) << 2 )
			| ( ( (instr_state >> COR_ACTIVE)     & 0x01 ) << 3 )
			);		
		}
	}
}