/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * fault.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * Self check functions and hardware error routines
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>

#include "fault.h"
#include "def.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  annunc_pattern[16];




/*
************************************************************************************************************************************************
Self check functions:
The MCU periodically checks the loss of lock signal of the Si5341 PLL synthesizer and displays an error code if a loss of lock is detected. 
*/

#define FAULTDET_PORT		PORTC
#define FAULTDET_PIN		PINC
#define FAULTDET_DDR		DDRC
#define PIN_LOL				PINC6

void fault_init(void)
{
	FAULTDET_DDR  &= ~( (1<<PIN_LOL) );														// Set PIN_LOL as input
	FAULTDET_PORT &= ~( (1<<PIN_LOL) );														// Do not use the internal pull up on PIN_LOL
}

void handle_error(uint8_t error_code)
{
	instr_error |= (1<<error_code);															// Set appropriate error flag in instr_error	
}

void clear_error(uint8_t error_code)
{
	instr_error &= ~( (1<<error_code) );													// Clear flag in instr_error
}


/*
************************************************************************************************************************************************
Testing for loss of lock of the PLL synthesizer:
This function is called periodically in the main function.
*/

void PLL_fault_test(void)
{
	if( !( FAULTDET_PIN & (1<<PIN_LOL) ) )													// Does the Si5341 indicate no loss of lock?
		clear_error(ERR_PLL_LOL);												
	else																					// Loss of lock detected, set error flag
		handle_error(ERR_PLL_LOL);
}


/*
************************************************************************************************************************************************
Testing the analog +5VA and -5VA supply rails:
This function is called upon startup of the receiver.
*/

void power_supply_test(void)																// Tests the analog +5VA and -5VA supply rails for proper voltage levels
{
	uint8_t i;
	uint8_t j;
	uint16_t v_inp_pos = 0;
	uint16_t v_inp_neg = 0;
	int16_t neg_supply_voltage;
		
	ADCSRA |= (1<<ADEN);																	// Enable ADC 
	ADCSRA &= ~( (1<<ADIE) );																// Disable ADC interrupt
	ADCSRA |= (1<<ADPS0) | (1<<ADPS1) | (1<<ADPS2);											// Select ADC prescaler to divide by 128 (= 62.5 kHz ADC clock)
	ADCSRA &= ~( (1<<ADATE) );																// ADC auto trigger disable
	ADMUX  &= ~( (1<<REFS0) | (1<<REFS1) );													// Use external reference voltage (Vref = 3.3V)
	ADMUX  &= ~( (1<<ADLAR) );																// ADC result right adjust
	ADMUX  &= ~( (1<<MUX0) | (1<<MUX1) | (1<<MUX2) | (1<<MUX3) | (1<<MUX4) );				// Clear ADC multiplex bits
	
	ADMUX |= (1<<MUX1);																		// Select ADC input 2 to measure positive supply voltage
	for( i = 0; i <= 3; i++)																// Take four measurements of supply voltage
	{
		ADCSRA |= (1<<ADSC);																// Start ADC conversion
		while( ADCSRA & (1<<ADSC) )															// Wait for ADC completion
			NOP();
		v_inp_pos += ADC;																	// Read ADC result
	}
	v_inp_pos /= 4;																			// Calculate mean value of four measurements
	
	ADMUX |= (1<<MUX0) | (1<<MUX1);															// Select ADC input 3 to measure negative supply voltage
	for( j = 0; j <= 3; j++ )																// Take four measurements of supply voltage
	{
		ADCSRA |= (1<<ADSC);																// Start ADC conversion
		while( ADCSRA & (1<<ADSC) )															// Wait for ADC completion
			NOP();
		v_inp_neg += ADC;																	// Read ADC result
	}
	v_inp_neg /= 4;																			// Calculate mean value of four measurements
	
	neg_supply_voltage = 3*v_inp_pos - 2*v_inp_neg;											// Calculate (magnitude of) negative input voltage (same scale as for positive voltage), as determined by voltage divider on A2 board (limits: lower=2.22V, upper=2.77)
	if( (v_inp_pos < 690) || (v_inp_pos > 860) )											// Check is positive voltage is not within limits (+/-10 percent)
		handle_error(ERR_POS_5V_ANALOG);
	else if( (neg_supply_voltage < 690) || (neg_supply_voltage > 860) )						// If positive voltage is within limits check if negative voltage is within limits
		handle_error(ERR_NEG_5V_ANALOG);
	
	ADCSRA &= ~( (1<<ADEN) );																// Disable ADC
}