/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * instr_control.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * Data output to the MCU ports
 */ 


#include "def.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>

#include "instr_control.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint32_t tuned_freq;
extern volatile uint8_t  cor_level;
extern volatile uint8_t  rssi_reading;




/*
************************************************************************************************************************************************
Instrument Control Output:
These routines output the instrument state to the appropriate MC ports.
*/

#define CONTR_PORT1			PORTB
#define CONTR_PORT2			PORTD
#define CONTR_PORT3			PORTG
#define CONTR_DDR1			DDRB
#define CONTR_DDR2			DDRD
#define CONTR_DDR3			DDRG
#define PORTPIN_SERDATA	    PORTB0
#define PORTPIN_DCLK		PORTB2
#define PORTPIN_STROBE		PORTB3
#define PORTPIN_COR			PORTD2
#define PORTPIN_STSLED1		PORTG0
#define PORTPIN_STSLED2		PORTG1


void contr_init(void)
{
	CONTR_DDR1 |= (1<<PORTPIN_SERDATA) | (1<<PORTPIN_DCLK) | (1<<PORTPIN_STROBE);			// Set port pins as output
	CONTR_DDR2 |= (1<<PORTPIN_COR);
	CONTR_DDR3 |= (1<<PORTPIN_STSLED1) | (1<<PORTPIN_STSLED2);
}

static void contr_process(void)																// Processes dependencies in instr_state
{
	cli();																					// Make local copies for evaluation of if statements to avoid timing issues
	uint32_t instr_state_loc = instr_state;
	uint8_t rssi_reading_loc = rssi_reading;
	uint8_t cor_level_loc    = cor_level;
	sei();
	
	// COR and mute
	if( instr_error )																		// Is an error flag set?
	{
		instr_state |= (1UL<<MUTE_RCVR);													// Mute receiver
		instr_state &= ~( (1UL<<COR_ACTIVE) );												// COR Active off
	}
	else if( (cor_level_loc >= MIN_COR_LEVEL) || (instr_state_loc & (1UL<<MGC)) )			// Is COR level set to min COR level or receiver is in MGC mode?
	{
		instr_state |= (1UL<<COR_ACTIVE);													// Then COR Active on
		if( ((1UL<<FM)|(1UL<<AM)|(1UL<<USB)|(1UL<<LSB)) & instr_state_loc )					// Is Mute not selected on front panel, i.e., one modulation mode is selected?
			instr_state &= ~( (1UL<<MUTE_RCVR) );											// Mute off
		else																				// Mute is selected on front panel
			instr_state |= (1UL<<MUTE_RCVR);												// Mute receiver
	}
	else if( cor_level_loc >= rssi_reading_loc )											// Is measured RSSI level above COR level?
	{
		instr_state |= (1UL<<COR_ACTIVE);													// COR Active on
		if( ((1UL<<FM)|(1UL<<AM)|(1UL<<USB)|(1UL<<LSB)) & instr_state )						// Is Mute not selected on front panel, i.e. one modulation mode is selected?
			instr_state &= ~( (1UL<<MUTE_RCVR) );											// Mute off
		else																				// Mute is selected on front panel
			instr_state |= (1UL<<MUTE_RCVR);												// Mute receiver	
	}
	else																					// No error and measured RSSI level is below COR level
	{
		instr_state |= (1UL<<MUTE_RCVR);													// Mute receiver
		instr_state &= ~( (1UL<<COR_ACTIVE) );												// COR Active off
	}
	
	// Frequency unit LEDs on front panel
	if( instr_error )																		// Is an error flag set?
		instr_state &= ~(1UL<<KHZ_LED), instr_state &= ~( (1UL<<MHZ_LED) );					// Then extinguish both LEDs
	else if( !(instr_state & (1UL<<SHIFT)) )												// Keyboard not shifted?
	{
		if( !(instr_state & (1UL<<BFO_TUNE)) )												// Receiver not in BFO tune mode (i.e. in set COR or main tuning mode)?
			instr_state |= (1UL<<MHZ_LED), instr_state &= ~( (1UL<<KHZ_LED) );				// Then light MHz LED and extinguish kHz LED
		else																				// Receiver is in BFO tune mode
			instr_state |= (1UL<<KHZ_LED), instr_state &= ~( (1UL<<MHZ_LED) );				// Then light kHz LED and extinguish MHz LED
	}
	else																					// Keyboard is shifted
	{
		if( instr_state & (1UL<<BFO_TUNE) )													// Receiver in BFO tune/entry mode?
			instr_state |= (1UL<<KHZ_LED), instr_state &= ~( (1UL<<MHZ_LED) );				// Then light kHz LED and extinguish MHz LED
		else if( instr_state & (1UL<<SET_CORLVL) )											// Receiver in set COR level mode?
			instr_state &= ~(1UL<<KHZ_LED), instr_state &= ~( (1UL<<MHZ_LED) );				// Then extinguish both LEDs
		else																				// Receiver in main frequency tune/entry mode
			instr_state |= (1UL<<MHZ_LED), instr_state &= ~( (1UL<<KHZ_LED) );				// Then light MHz LED and extinguish kHz LED
	}
}

static void write_portextend(uint8_t *write_buffer, uint8_t len)							// Bit bang len bytes in write_buffer to port extender shift register
{
	int8_t i;
	int8_t j;
	CONTR_PORT1 &= ~( (1<<PORTPIN_STROBE) );												// Strobe line low
	for( i = len-1; i >= 0; i-- )
	{
		for( j = 7; j >= 0; j-- )
		{
			CONTR_PORT1 &= ~(1<<PORTPIN_DCLK);												// Clock line low
			_delay_us(1);																	// Wait
			if( (1<<j) & (*(write_buffer + i)) )											// Next bit 1?
				CONTR_PORT1 |= (1<<PORTPIN_SERDATA);										// Serial data line high
			else																			// Next bit 0
				CONTR_PORT1 &= ~( (1<<PORTPIN_SERDATA) );									// Serial data line low
			_delay_us(1);																	// Wait
			CONTR_PORT1 |= (1<<PORTPIN_DCLK);												// Clock line high
			_delay_us(1);																	// Wait
		}
	}
	CONTR_PORT1 &= ~(1<<PORTPIN_DCLK);														// Clock line low
	CONTR_PORT1 |= (1<<PORTPIN_STROBE);														// Strobe line high
	_delay_us(1);																			// Wait
	CONTR_PORT1 &= ~( (1<<PORTPIN_STROBE) );												// Strobe line low
}

static void create_buffer(uint8_t *buffer)
{
	// Process bandwidth select
	if( instr_state & (1UL<<BW1) )
		*(buffer + 0) |= (1<<PEXT_BW1);														// Select BW1
	else if( instr_state & (1UL<<BW2) )
		*(buffer + 0) |= (1<<PEXT_BW2);														// Select BW2
	else if( instr_state & (1UL<<BW3) )
		*(buffer + 0) |= (1<<PEXT_BW3);														// Select BW3
	else if( instr_state & (1UL<<BW4) )
		*(buffer + 0) |= (1<<PEXT_BW4);														// Select BW4
	// Process mute and detection mode
	if( !(instr_state & (1UL<<MUTE_RCVR)) )													// Receiver not in muted state? (if muted, no modulation mode must be set)
	{
		if ( instr_state & (1UL<<AM) )
			*(buffer + 0) |= (1<<PEXT_AM);													// Select AM detection
		else if( instr_state & ((1UL<<USB)|(1UL<<LSB)) )
			*(buffer + 0) |= (1<<PEXT_SSB);													// Select SSB detection
		else if( instr_state & (1UL<<FM) )
			*(buffer + 0) |= (1<<PEXT_FM);													// Select FM detection		
	}
	if( instr_state & (1UL<<LSB) )
		*(buffer + 0) |= (1<<PEXT_USB);														// Set USB/LSB line high
	// Process AGC
	if( !(instr_state & (1UL<<MGC)) )
		*(buffer + 1) |= (1<<PEXT_AGC_AUTO);												// AGC auto line high when receiver is not set to MGC
	if( instr_state & (1UL<<AGC_FAST) )
		*(buffer + 1) |= (1<<PEXT_AGC_FAST);												// Select AGC fast
	else if( instr_state & (1UL<<AGC_SLOW) )
		*(buffer + 1) |= (1<<PEXT_AGC_SLOW);												// Select AGC slow
	else if( instr_state & (1UL<<AGC_HANG) )
		*(buffer + 1) |= (1<<PEXT_AGC_HANG);												// Select AGC hang
	// Control lines
	if( !(instr_state & (1UL<<ATT_IN)) )
		*(buffer + 1) |= (1<<PEXT_ATT_OUT);													// Select attenuator
	if( instr_state & (1UL<<PREAMP_IN) )
		*(buffer + 1) |= (1<<PEXT_PREAMP_ON);												// Select preamp on
	if( !(instr_state & (1<<SPEAKER_ON)) )
		*(buffer + 1) |= (1<<PEXT_SPKR_OFF);												// Select speaker off
	// Process preselector
	if( !(instr_state & (1UL<<PRESEL_IN)) )													// Is preselector not selected on front panel?
		*(buffer + 2) |= (1<<PEXT_PS_OFF);													// Then select preselector bypass (preselector out)
	else																					// Preselector is selected in on front panel
	{
		if( tuned_freq <= PRESEL_BL_1 )
			*(buffer + 2) |= (1<<PEXT_PS_BD1);												// Select band 1
		else if( tuned_freq <= PRESEL_BL_2 )
			*(buffer + 2) |= (1<<PEXT_PS_BD2);												// Select band 2
		else if( tuned_freq <= PRESEL_BL_3 )
			*(buffer + 2) |= (1<<PEXT_PS_BD3);												// Select band 3
		else if( tuned_freq <= PRESEL_BL_4 )
			*(buffer + 2) |= (1<<PEXT_PS_BD4);												// Select band 4
		else if( tuned_freq <= PRESEL_BL_5 )
			*(buffer + 2) |= (1<<PEXT_PS_BD5);												// Select band 5
		else
			*(buffer + 2) |= (1<<PEXT_PS_BD6);												// Select band 6		
	}
}

void contr_out(void)																		// Outputs instr_state and preselector state to the receiver
{
	static uint32_t instr_state_old;														// Save instrument state for detection of changes
	static uint32_t tuned_freq_old;															// Saved tuned frequency value for detection of changes
	uint8_t portextend_buffer[3] = {0};														// Buffer that is called by reference and outputted to I/O port extender on A2 board
	uint8_t flag;																			// Flag signifying that AGC has been switched to slow or hang (in that case pull "AGC Fast" line high for 0.1 s to quickly reach steady state in AGC control loop)
	
	contr_process();																		// Process dependencies in instr_state
	
	if( tuned_freq != tuned_freq_old || instr_state != instr_state_old )					// Check if instr_state or tuned_freq has changed
	{
		flag = ( ((instr_state^instr_state_old) & ((1UL<<AGC_SLOW)|(1UL<<AGC_HANG))) && (instr_state & ((1UL<<AGC_SLOW)|(1UL<<AGC_HANG))) ) ? 1 : 0;
		if(!flag)																			// Has receiver not been set to slow or hanging AGC?
		{
			create_buffer( portextend_buffer );
			write_portextend( portextend_buffer, 3 );			
		}
		else																				// Receiver has been set to slow or hanging AGC
		{
			create_buffer( portextend_buffer );												// Create buffer to be sent to port extender 
			portextend_buffer[1] |= (1<<PEXT_AGC_FAST);										// Modify buffer to select AGC fast in order to quickly reach steady state in AGC control loop
			write_portextend( portextend_buffer, 3 );										// Output modified buffer to port extender shift register
			_delay_ms(100);																	// Wait 100 msec
			memset( portextend_buffer, 0, 3 );												// Clear buffer
			create_buffer( portextend_buffer );												// Create buffer again
			write_portextend( portextend_buffer, 3 );										// Output buffer				
		}
		
		if( instr_state & (1UL<<COR_ACTIVE) )												// Output rear panel COR interface
			CONTR_PORT2 |= (1<<PORTPIN_COR);
		else
			CONTR_PORT2 &= ~( (1<<PORTPIN_COR) );
		// Update old instr_state and frequency
		instr_state_old = instr_state;
		tuned_freq_old = tuned_freq;
	}
}

void status_LED_out(void)																	// Toggles the status LEDs on A2 board after a predetermined number of calls
{
	static uint16_t status_count;
	if( status_count < 1024 )
		++status_count;
	else
	{
		status_count = 0;
		CONTR_PORT3 ^= (1<<PORTPIN_STSLED1);
		if( CONTR_PORT3 & (1<<PORTPIN_STSLED1) )
			CONTR_PORT3 &= ~(1<<PORTPIN_STSLED2);
		else
			CONTR_PORT3 |= (1<<PORTPIN_STSLED2);
	}
}