/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * keyboard.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * Keyboard poll and updating of instr_state and other global variables
 */ 


#include "def.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "keyboard.h"
#include "fault.h"

extern volatile uint32_t instr_state;
extern volatile uint8_t  instr_error;
extern volatile uint8_t  last_key;
extern volatile uint32_t tuned_freq;
extern volatile int16_t  bfo_offset; 
extern volatile struct 
	{
	uint8_t digit[8];
	uint8_t decpoint_pos;
	uint8_t decpoint_flag;
	uint8_t	digit_count;
	uint8_t sign;
	} scratchpad;




/*
************************************************************************************************************************************************
Keyboard Poll:
The keyboard is polled every 10ms in the ISR corresponding to Timer2 Compare Match. If a key is found depressed, key_scan returns the
appropriate key number. The function key_scan is called by the ISR, and its result is stored in last_key. This variable is read by read_keyboard, 
which also clears it.
*/

#define ROWPORT					PORTC
#define ROWPIN					PINC
#define ROWDDR					DDRC
#define COLPORT					PORTD
#define COLPIN					PIND
#define COLDDR					DDRD
#define ROW1					PORTC0
#define ROW2					PORTC1
#define ROW3					PORTC2
#define ROW4					PORTC3
#define ROW5					PORTC4
#define COL1					PORTD3
#define COL2					PORTD4
#define COL3					PORTD5
#define COL4					PORTD6
#define COL5					PORTD7

static const uint32_t pow10[ ] = {1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000};	// Powers of 10:


void key_init(void)
{
	ROWDDR  |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Set rows as output for scan of columns
	COLDDR  &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );				// Set columns as input for scan of columns
	ROWPORT &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Rows low
	COLPORT |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );				// Cols pull-up on

	TCCR2 |= (1<<CS20) | (1<<CS22);															// Timer2 prescaler set to divide by 1024
	TCCR2 |= (1<<WGM21);																	// Clear Timer2 on compare
	TIMSK |= (1<<OCIE2);																	// Timer2 output compare match interrupt enable
	OCR2 = 156 - 1;																			// Compare register, set to 155 = 156-1, interrupt approx every 10ms
}

static uint8_t key_scan(void)																// Returns key number (keys are numbered from 1 to 25), and 0 for no key pressed, should be called every 10ms
{
	static uint8_t scan_dir;
	static uint8_t row;
	static uint8_t col;
	uint8_t key = 0;
	
	static uint8_t oldrow;
	static uint8_t oldcol;
	static uint8_t debounce;

	if( scan_dir == 0 )
	{
		if( ~COLPIN & (1 << COL1) )															// Read col
			col = 6;
		if( ~COLPIN & (1 << COL2) )
			col = 11;
		if( ~COLPIN & (1 << COL3) )
			col = 16;
		if( ~COLPIN & (1 << COL4) )
			col = 21;
		if( ~COLPIN & (1 << COL5) )
			col = 26;
		
		ROWDDR  &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Set rows as input for scan of rows
		COLDDR  |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );			// Set cols as output for scan of rows
		COLPORT &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );			// Columns low for scan of rows
		ROWPORT |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Rows pull-up on for scan of rows
			
		scan_dir = 1;																		// Toggle scan direction; scan_dir=0 means scan columns, scan_dir=1 means scan rows
		return 0;
	}
	else
	{
		if( ~ROWPIN & (1<<ROW1) )															// Read row
			row = 5;
		if( ~ROWPIN & (1<<ROW2) )
			row = 4;
		if( ~ROWPIN & (1<<ROW3) )
			row = 3;
		if( ~ROWPIN & (1<<ROW4) )
			row = 2;
		if( ~ROWPIN & (1<<ROW5) )
			row = 1;
			
		ROWDDR  |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Set rows as output for scan of columns
		COLDDR  &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );			// Set columns as input for scan of columns
		ROWPORT &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );			// Rows low for scan of columns
		COLPORT |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );			// Cols pull-up on for scan of columns
					
		scan_dir = 0;																		// Toggle scan direction
		key = col - row;																	// Determine key number, yields zero if no key is pressed
		
		if( oldrow != row  || oldcol != col )												// Debouncing algorithm
		{
			oldrow = row;
			oldcol = col;
			debounce = 2;
		}		
		row = 0;
		col = 0;		
		
		if( key != 0 && debounce != 0 && --debounce == 0 )									// Old key equals key for four calls (= 40 ms)
			return key;
		return 0;	
	}
}

ISR(TIMER2_COMP_vect)																		// Interrupt service routine for Timer2 compare match
{
	uint8_t current_key = key_scan();
	if( (current_key != 0) && (last_key == 0) )												// Ignore further key presses if last_key has not been processed and cleared by read_keyboard
	{
		last_key = current_key;
	}
}




/*
************************************************************************************************************************************************
Read Data from last_key and update instr_state appropriately, or read numeric keypad when Shift is selected and store entries in scratchpad and evaluate scratchpad
*/

static uint8_t get_digit(uint8_t key_code)													// Converts key code on numeric keypad to corresponding digit
{
	switch( key_code )
	{
		case KEY_BW4:
			return 0;
			break;
		case KEY_FM:
			return 1;
			break;
		case KEY_USB:
			return 2;
			break;
		case KEY_LSB:
			return 3;
			break;
		case KEY_MGC:
			return 4;
			break;
		case KEY_HANG:
			return 5;
			break;
		case KEY_SLOW:
			return 6;
			break;
		case KEY_BW1:
			return 7;
			break;
		case KEY_BW2:
			return 8;
			break;
		case KEY_BW3:
			return 9;
			break;
		case KEY_FAST:																		// Decimal point
			return 100;
			break;
		case KEY_100KHZ:																	// Change sign
			return 101;
			break;
		case KEY_AM:																		// Exe key
			return 102;
			break;
		default:																			// Other key (not on numeric keypad)
			return 200;
			break;
	}
}

static void clear_scratchpad(void)															// Clears scratchpad (set everything to 0)
{
	uint8_t j;
	for( j = 0; j < sizeof(scratchpad.digit) / sizeof(scratchpad.digit[0]); j++ )
		scratchpad.digit[j] = 0;
	scratchpad.decpoint_flag = 0;
	scratchpad.decpoint_pos = 0;
	scratchpad.digit_count = 0;
	scratchpad.sign = 0;
}

static void process_scratchpad(void)														// Processes numeric keypad entry stored in scratchpad and write result to appropriate global variable (tuned_freq or bfo_offset)
{																							// This function is to be called when Exe key is pressed
	uint8_t i;
	int32_t number_entry = 0;
	uint8_t number_of_digits;

	if( (scratchpad.digit_count != 0) && (scratchpad.decpoint_pos <= 2) )					// Is there an entry in scratchpad and are less than two digits in front of the decimal point?
	{
		if( scratchpad.digit_count - scratchpad.decpoint_pos > 6 )							// More than six decimals were entered?
			number_of_digits = 6 + scratchpad.decpoint_pos;									// Then discard all but the first six decimals
		else																				// Six or less decimals were entered
			number_of_digits = scratchpad.digit_count;										// Process all entered digits
		for( i = 0; i < number_of_digits; i++ )												// Calculate entered int32_t number_entry from digits stored in scratchpad
			number_entry += ( scratchpad.digit[i] * pow10[5 - i + scratchpad.decpoint_pos] );
		if( scratchpad.sign )																// Negative sign in scratchpad?
			number_entry = -number_entry;			
		if( instr_state & (1UL<<BFO_TUNE) )													// Receiver in BFO tune mode?
		{
			number_entry /= 1000;															// Divide by 1000 because BFO entry is in kHz and not MHz
			if( (number_entry <= MAX_BFO_OFFS) && (number_entry >= MIN_BFO_OFFS) )			// Check if entry is within permissible bounds
				bfo_offset = (int16_t)number_entry;											// Write entry to global
		}
		else if( instr_state & (1UL<<SET_CORLVL) )											// Receiver is in set COR level mode?
			NOP();																			// Do nothing (discard entry) as COR level can currently only be set by encoder
		else																				// Receiver is in frequency tune mode
		{
			if( (number_entry <= MAX_FREQ) && (number_entry >= MIN_FREQ) )					// Check if entry is within permissible bounds, else discard entry
				tuned_freq = (uint32_t)number_entry;										// Write entry to global
		}	
	}
	instr_state &= ~(1UL<<SHIFT);															// Un-shift keyboard
	clear_scratchpad();																		// Clear scratchpad
}

static void numpad(uint8_t key_code)														// Evaluates numeric keypad entries and stores entered digits in scratchpad,
{																							// and processes scratchpad when Exe key is pressed
	uint8_t digit = get_digit( key_code );													// Convert key_code to digit
	uint8_t max_digits;																		// Maximum number of digits that can be entered via keypad
	
	if( !(instr_state & (1UL<<BFO_TUNE)) )													// Receiver not in BFO tune mode?
		max_digits = 8;																		// Then eight digits can be entered
	else
		max_digits = 7;																		// Else only seven digits can be entered to leave room for the sign

	if( (digit < 100) && (scratchpad.digit_count < max_digits) )							// Number key pressed and buffer not yet full?
	{
		scratchpad.digit[scratchpad.digit_count] = digit;									// Store digit in appropriate register in scratchpad
		scratchpad.digit_count++;															// Increment digit count
		if( !scratchpad.decpoint_flag )														// Decimal point key was not pressed before?
			scratchpad.decpoint_pos++;														// Then advance decimal point position indicator by one as another digit is entered
	}
	if( digit == 100 )																		// Decimal point pressed?
	{
		scratchpad.decpoint_flag = 1;														// Set flag
		if( scratchpad.digit_count == 0 )													// First key pressed is decimal point?
		{
			scratchpad.digit[0] = 0;														// Then make first digit null and set decimal point position to 1.
			scratchpad.decpoint_pos = 1;
			scratchpad.digit_count = 1; 	
		}
	}
	if( (digit == 101) && ( (instr_state & (1UL<<BFO_TUNE)) != 0 ) )						// Change sign key pressed and receiver in BFO tune mode?
		scratchpad.sign ^= 1;																// Then toggle sign in scratchpad
	if( digit == 102 )																		// Exe key pressed? (ignore all other keys)
	{
		if( scratchpad.digit_count == 0 )													// First key pressed is Exe, or Exe is pressed directly after change sign?
		{
			scratchpad.digit[0] = 0;														// Then make first digit null and advance digit count by one
			scratchpad.digit_count = 1;
			scratchpad.decpoint_pos = 1;
		}
		process_scratchpad();																// Process scratchpad and write contents, depending on instr_state, to the appropriate globals																	
	}
}

void read_keyboard(void)																	// Updates instr_state from last_key or reads numeric keypad when Shift is selected
{
	uint8_t last_key_loc;
	cli();
	last_key_loc = last_key;																// Make local copy of last_key
	last_key = 0;																			// Clear last_key
	sei();
	
	if( last_key_loc && !instr_error )														// Any key pressed and no error flag set?
	{
		if( !(instr_state & (1UL<<SHIFT)) )													// Keyboard not shifted?
		{
			switch( last_key_loc )															// Update instr_state according to key pressed
			{
				case KEY_AM:
					if( !(instr_state & (1UL<<AM)) )
					{
						instr_state |= (1UL<<AGC_FAST), instr_state &= ~((1UL<<AGC_SLOW)|(1UL<<AGC_HANG)|(1UL<<MGC));
						instr_state |= (1UL<<BW3), instr_state &= ~((1UL<<BW1)|(1UL<<BW2)|(1UL<<BW4));
					}
					instr_state ^= (1UL<<AM), instr_state &= ~((1UL<<USB)|(1UL<<LSB)|(1UL<<FM));				
					break;
				case KEY_USB:
					if( !(instr_state & (1UL<<USB)) )
					{
						instr_state |= (1UL<<AGC_HANG), instr_state &= ~((1UL<<AGC_FAST)|(1UL<<AGC_SLOW)|(1UL<<MGC));
						instr_state |= (1UL<<BW2), instr_state &= ~((1UL<<BW1)|(1UL<<BW3)|(1UL<<BW4));					
					}
					instr_state ^= (1UL<<USB), instr_state &= ~((1UL<<AM)|(1UL<<LSB)|(1UL<<FM));
					break;
				case KEY_LSB:
					if( !(instr_state & (1UL<<LSB)) )
					{
						instr_state |= (1UL<<AGC_HANG), instr_state &= ~((1UL<<AGC_FAST)|(1UL<<AGC_SLOW)|(1UL<<MGC));
						instr_state |= (1UL<<BW2), instr_state &= ~((1UL<<BW1)|(1UL<<BW3)|(1UL<<BW4));
					}
					instr_state ^= (1UL<<LSB), instr_state &= ~((1UL<<AM)|(1UL<<USB)|(1UL<<FM));
					break;
				case KEY_FM:
					if( !(instr_state & (1UL<<FM)) )
					{
						instr_state |= (1UL<<AGC_FAST), instr_state &= ~((1UL<<AGC_SLOW)|(1UL<<AGC_HANG)|(1UL<<MGC));
						instr_state |= (1UL<<BW4), instr_state &= ~((1UL<<BW1)|(1UL<<BW2)|(1UL<<BW3));
					}
					instr_state ^= (1UL<<FM), instr_state &= ~((1UL<<AM)|(1UL<<USB)|(1UL<<LSB));
					break;
				case KEY_FAST:
					instr_state |= (1UL<<AGC_FAST), instr_state &= ~((1UL<<AGC_SLOW)|(1UL<<AGC_HANG)|(1UL<<MGC));
					break;
				case KEY_SLOW:
					instr_state |= (1UL<<AGC_SLOW), instr_state &= ~((1UL<<AGC_FAST)|(1UL<<AGC_HANG)|(1UL<<MGC));
					break;
				case KEY_HANG:
					instr_state |= (1UL<<AGC_HANG), instr_state &= ~((1UL<<AGC_FAST)|(1UL<<AGC_SLOW)|(1UL<<MGC));
					break;
				case KEY_MGC:
					instr_state |= (1UL<<MGC), instr_state &= ~((1UL<<AGC_FAST)|(1UL<<AGC_SLOW)|(1UL<<AGC_HANG));
					break;
				case KEY_BW1:
					instr_state |= (1UL<<BW1), instr_state &= ~((1UL<<BW2)|(1UL<<BW3)|(1UL<<BW4));
					break;
				case KEY_BW2:
					instr_state |= (1UL<<BW2), instr_state &= ~((1UL<<BW1)|(1UL<<BW3)|(1UL<<BW4));
					break;
				case KEY_BW3:
					instr_state |= (1UL<<BW3), instr_state &= ~((1UL<<BW1)|(1UL<<BW2)|(1UL<<BW4));
					break;
				case KEY_BW4:
					instr_state |= (1UL<<BW4), instr_state &= ~((1UL<<BW1)|(1UL<<BW2)|(1UL<<BW3));
					break;
				case KEY_SETCOR:
					instr_state ^= (1UL<<SET_CORLVL), instr_state &= ~(1UL<<BFO_TUNE);
					break;
				case KEY_BFOTUNE:
					instr_state ^= (1UL<<BFO_TUNE), instr_state &= ~(1UL<<SET_CORLVL);
					break;
				case KEY_ATT:
					instr_state ^= (1UL<<ATT_IN);
					break;
				case KEY_PRESEL:
					instr_state ^= (1UL<<PRESEL_IN);
					break;
				case KEY_PREAMP:
					instr_state ^= (1UL<<PREAMP_IN);
					break;
				case KEY_SPKR:
					instr_state ^= (1UL<<SPEAKER_ON);
					break;
			}
			// The evaluation of the tuning speed key group depends on whether or not the receiver is in BFO tune mode
			if( !( instr_state & (1UL<<BFO_TUNE) ) )										// Is receiver not in BFO tuning mode?
			{
				switch( last_key_loc )														// Evaluate tuning speed key group
				{
					case KEY_1HZ:
						instr_state ^= (1UL<<TUNE_1HZ), instr_state &= ~((1UL<<TUNE_10HZ)|(1UL<<TUNE_100HZ)|(1UL<<TUNE_1KHZ)|(1UL<<TUNE_10KHZ)|(1UL<<TUNE_100KHZ));
						break;
					case KEY_10HZ:
						instr_state ^= (1UL<<TUNE_10HZ), instr_state &= ~((1UL<<TUNE_1HZ)|(1UL<<TUNE_100HZ)|(1UL<<TUNE_1KHZ)|(1UL<<TUNE_10KHZ)|(1UL<<TUNE_100KHZ));
						break;
					case KEY_100HZ:
						instr_state ^= (1UL<<TUNE_100HZ), instr_state &= ~((1UL<<TUNE_1HZ)|(1UL<<TUNE_10HZ)|(1UL<<TUNE_1KHZ)|(1UL<<TUNE_10KHZ)|(1UL<<TUNE_100KHZ));
						break;
					case KEY_1KHZ:
						instr_state ^= (1UL<<TUNE_1KHZ), instr_state &= ~((1UL<<TUNE_1HZ)|(1UL<<TUNE_10HZ)|(1UL<<TUNE_100HZ)|(1UL<<TUNE_10KHZ)|(1UL<<TUNE_100KHZ));
						break;
					case KEY_10KHZ:
						instr_state ^= (1UL<<TUNE_10KHZ), instr_state &= ~((1UL<<TUNE_1HZ)|(1UL<<TUNE_10HZ)|(1UL<<TUNE_100HZ)|(1UL<<TUNE_1KHZ)|(1UL<<TUNE_100KHZ));
						break;
					case KEY_100KHZ:
						instr_state ^= (1UL<<TUNE_100KHZ), instr_state &= ~((1UL<<TUNE_1HZ)|(1UL<<TUNE_10HZ)|(1UL<<TUNE_100HZ)|(1UL<<TUNE_1KHZ)|(1UL<<TUNE_10KHZ));
						break;	
				}
			}
			else																			// Receiver is in BFO tuning mode
			{
				switch( last_key_loc )														// Evaluate tuning speed key group
				{
					case KEY_1HZ:
						instr_state |= (1UL<<BFOTUNE_1HZ), instr_state &= ~((1UL<<BFOTUNE_10HZ)|(1UL<<BFOTUNE_100HZ));
						break;
					case KEY_10HZ:
						instr_state |= (1UL<<BFOTUNE_10HZ), instr_state &= ~((1UL<<BFOTUNE_1HZ)|(1UL<<BFOTUNE_100HZ));
						break;
					case KEY_100HZ:
						instr_state |= (1UL<<BFOTUNE_100HZ), instr_state &= ~((1UL<<BFOTUNE_10HZ)|(1UL<<BFOTUNE_1HZ));
						break;				
				}													
			}
		}
		else																				// Keyboard shifted
		{
			numpad(last_key_loc);															// Pass depressed key on to numeric keypad evaluation		
		}
																						
		// Now evaluate shift key																
		if( last_key_loc == KEY_SHIFT )														// Shift key pressed?
		{
			instr_state ^= (1UL<<SHIFT);													// Then toggle shift state
			clear_scratchpad();																// Clear scratchpad		
		}
	}
}




/*
************************************************************************************************************************************************
Test for a stuck key at startup
*/

void stuck_key_test(void)
{
	ROWDDR  &= ~( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Set rows as input for scan of rows
	COLDDR  |=  ( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );				// Set cols as output for scan of rows
	COLPORT &= ~( (1<<COL1) | (1<<COL2) | (1<<COL3) | (1<<COL4) | (1<<COL5) );				// Cols low for scan of rows
	ROWPORT |=  ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) );				// Rows pull-up on for scan of rows	
	_delay_ms(20);
	if( ~ROWPIN & ( (1<<ROW1) | (1<<ROW2) | (1<<ROW3) | (1<<ROW4) | (1<<ROW5) ) )			// Any row not high?
		handle_error(ERR_STUCK_KEY);
}