/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * si5341.c
 *
 * Version 1.0
 *
 * Created: 01.07.2018
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: HF Receiver Mark 2 Firmware
 *
 * Controls the Si5341 PLL synthesizer for LO1, LO2, and the BFO, via the TWI bus
 */ 


#include "def.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "si5341.h"
#include "twi.h"

extern volatile uint32_t tuned_freq;
extern volatile int16_t  bfo_offset;




/*
************************************************************************************************************************************************
Reading and Writing to Single and to Consecutive Registers of the Si5341:
*/

#define SI5341_ADDRESS		0b01110111														// 7 bit Si5341 TWI bus address as determined by hardware

static void si5341_write_register(uint8_t addr, uint8_t target_register, uint8_t val)
{
	TWI_start();																			// Send start condition
	TWI_write( (addr<<1) | 0x00 );															// Send slave address and request write
	TWI_write(target_register);																// Send Si5341 register address
	TWI_write(val);																			// Send register content
	TWI_stop();																				// Send stop condition
}

static void si5341_write_consec(uint8_t addr, uint8_t start_addr_register, uint8_t num, uint8_t *write_buffer)
{
	uint8_t i;
	TWI_start();
	TWI_write( (addr<<1) | 0x00 );
	TWI_write(start_addr_register);
	for( i = 0; i < num; ++i )
	{
		TWI_write( *(write_buffer + i) );
	}
	TWI_stop();
}

static uint8_t si5341_read_register(uint8_t addr, uint8_t register_addr)
{
	uint8_t val;
	TWI_start();																			// Send start condition
		TWI_write( (addr<<1) | 0x00 );														// Send slave address and request write
		TWI_write(register_addr);															// Send Si5341 register address
		TWI_stop();																			// Send stop condition
		TWI_start();																		// Send start condition
		TWI_write( (addr<<1) | 0x01 );														// Send slave address and request read
		val = TWI_read_NACK();																// Read from TWI, send no Acknowledge
	TWI_stop();																				// Send stop condition
	return val;
}

static void si5341_read_consec(uint8_t addr, uint8_t start_addr_register, uint8_t num, uint8_t *read_buffer)
{
	uint8_t i;
	TWI_start();
		TWI_write( (addr<<1) | 0x00 );
		TWI_write(start_addr_register);
		TWI_stop();
		TWI_start();
		TWI_write( (addr<<1) | 0x01 );
		for( i = 0; i < num - 1; ++i )
		{
			*( read_buffer + i ) = TWI_read_ACK();
		}
		*( read_buffer + num - 1 ) = TWI_read_NACK();
	TWI_stop();
}

static void set_page(uint8_t page)
{
	static uint8_t last_page = 255;
	if( page != last_page )
	{
		si5341_write_register(SI5341_ADDRESS, 0x01, page);
		last_page = page;
	}
}




/*
************************************************************************************************************************************************
Initial Setup of the Si5341 PLL Frequency Synthesizer Device.
*/

#define SI5341_PORT			PORTC
#define SI5341_PORT_DDR		DDRC
#define PIN_DCO_RESET		PINC5

void si5341_init(void)
{
	SI5341_PORT_DDR |= (1<<PIN_DCO_RESET);													// Set DCO_RESET pin as output
	SI5341_PORT &= ~( (1<<PIN_DCO_RESET) );
	_delay_ms(100);
	SI5341_PORT |= (1<<PIN_DCO_RESET);														// Resets the Si5341
	
	set_page(0);
	si5341_write_register(SI5341_ADDRESS, 0x1E, 0b00000010);								// Perform hard reset
	_delay_ms(250);
	si5341_write_register(SI5341_ADDRESS, 0x1E, 0b00000000);								// Clear hard reset bit, low power mode disable
	si5341_write_register(SI5341_ADDRESS, 0x21, 0b00000110);								// Select pin controlled clock input selection, and select XAXB input as reference clock for PLL
	si5341_write_register(SI5341_ADDRESS, 0x2C, 0b00010000);								// Disable LOS on XAXB input
	
	set_page(9);
	si5341_write_register(SI5341_ADDRESS, 0x0E, 0b00000011);								// Enable external clock on XAXB input, power up buffer circuitry of XAXB input
	si5341_write_register(SI5341_ADDRESS, 0x1C, 0b00000100);								// Zero delay mode not enabled
	si5341_write_register(SI5341_ADDRESS, 0x43, 0b00000001);								// Select 3.3V thresholds for external I/O connections
	si5341_write_register(SI5341_ADDRESS, 0x49, 0b00000000);								// Disable IN0, IN1, IN2, IN3 clock inputs
	si5341_write_register(SI5341_ADDRESS, 0x4A, 0b00000000);								// Disables clock routing from IN0, IN1, IN2, IN3 to the phase detector
	
	set_page(0x0A);
	si5341_write_register(SI5341_ADDRESS, 0x03, 0b00001111);								// Enable multisynth dividers 0, 1, 2, 3
	si5341_write_register(SI5341_ADDRESS, 0x04, 0b00000000);								// Multisynth divider phase interpolator bypass disabled, multisynth divider used with fractional division ratio
	si5341_write_register(SI5341_ADDRESS, 0x05, 0b00001111);								// Power up multisynth dividers 0, 1, 2, 3
	
	set_page(0x0B);
	si5341_write_register(SI5341_ADDRESS, 0x4A, 0b00000000);								// Divider clock disables set as required
		
	set_page(1);
	si5341_write_register(SI5341_ADDRESS, 0x02, 0x01);										// OUTALL_DISABLE_LOW = 1, no output drivers are disabled by this variable
	si5341_write_register(SI5341_ADDRESS, 0x45, 0b00000000);								// OUT_PDN_ALL = 0, no output driver powered down by this signal (no effect)
	
	// Setup for Output 0 (LO1):
	si5341_write_register(SI5341_ADDRESS, 0x08, 0b00000110);								// Enable Output 0, power up output driver, force R0 divider to divide by 2
	si5341_write_register(SI5341_ADDRESS, 0x09, 0b11000001);								// Set Output 0 to normal differential (LVDS), disable output sync, disable in low state
	si5341_write_register(SI5341_ADDRESS, 0x0A, 0b00111011);								// Set LVDS common mode voltage to 2V and differential amplitude to 450 mV (cf. Table 6.4 in Si5341 Family Reference Manual and Table 16.1 in the Appendix)
	si5341_write_register(SI5341_ADDRESS, 0x0B, 0b00000000);								// Output 0 mux connected to N divider 0, no inversion on output (i.e. complementary)
	
	// Setup for Output 5 (BFO):
	si5341_write_register(SI5341_ADDRESS, 0x21, 0b00000110);								// Enable Output 5, power up output driver, force R4 divider to divide by 2
	si5341_write_register(SI5341_ADDRESS, 0x22, 0b11000100);								// Set Output 4 to LVCMOS, Z_out=22 Ohms, disable output sync, disable in low state
	si5341_write_register(SI5341_ADDRESS, 0x24, 0b00000001);								// Output 5 mux connected to N divider 1, no inversion on output
	
	// Setup for Output 6 (LO3/FM-LO):
	si5341_write_register(SI5341_ADDRESS, 0x26, 0b00000110);								// Enable Output 6, power up output driver, force R5 divider to divide by 2
	si5341_write_register(SI5341_ADDRESS, 0x27, 0b11000100);								// Set Output 5 to LVCMOS, Z_out=22 Ohms, disable output sync, disable in low state
	si5341_write_register(SI5341_ADDRESS, 0x29, 0b00000010);								// Output 6 mux connected to N divider 2, no inversion on output

	// Setup for Output 9 (LO2):
	si5341_write_register(SI5341_ADDRESS, 0x3A, 0b00000110);								// Enable Output 9, power up output driver, force R9 divider to divide by 2
	si5341_write_register(SI5341_ADDRESS, 0x3B, 0b11000100);								// Set Output 0 to LVCMOS, Z_out=22 Ohms, disable output sync, disable in low state
	si5341_write_register(SI5341_ADDRESS, 0x3D, 0b00000011);								// Output 9 mux connected to N divider 3, no inversion on output
	
	// Disable and power down all other output drivers
	si5341_write_register(SI5341_ADDRESS, 0x0D, 0x01);										// Output 1
	si5341_write_register(SI5341_ADDRESS, 0x12, 0x01);										// Output 2
	si5341_write_register(SI5341_ADDRESS, 0x17, 0x01);										// Output 3
	si5341_write_register(SI5341_ADDRESS, 0x1C, 0x01);										// Output 4
	si5341_write_register(SI5341_ADDRESS, 0x2B, 0x01);										// Output 7
	si5341_write_register(SI5341_ADDRESS, 0x30, 0x01);										// Output 8
	
	set_page(2);
	si5341_write_register(SI5341_ADDRESS, 0x02, 0);											// Set XAXB frequency adjust value to zero
	si5341_write_register(SI5341_ADDRESS, 0x03, 0);
	si5341_write_register(SI5341_ADDRESS, 0x04, 0);
	si5341_write_register(SI5341_ADDRESS, 0x05, 0);
	
	si5341_write_register(SI5341_ADDRESS, 0x06, 0);											// Set the divider on XAXB input to divide by one (no update bit for this divider)
	
	// PLL setup for f_VCO = 14 GHz
	si5341_write_register(SI5341_ADDRESS, 0x35, 0x5E);										// Set main PLL M divider numerator value to 350, since VCO is to run at 14.0 GHz (reference frequency is 40 MHz)
	si5341_write_register(SI5341_ADDRESS, 0x36, 0x01);
	si5341_write_register(SI5341_ADDRESS, 0x37, 0);
	si5341_write_register(SI5341_ADDRESS, 0x38, 0);
	si5341_write_register(SI5341_ADDRESS, 0x39, 0);
	si5341_write_register(SI5341_ADDRESS, 0x3A, 0);
	
	si5341_write_register(SI5341_ADDRESS, 0x3B, 1);											// Set main PLL M divider denominator value to 1
	si5341_write_register(SI5341_ADDRESS, 0x3C, 0);
	si5341_write_register(SI5341_ADDRESS, 0x3D, 0);
	si5341_write_register(SI5341_ADDRESS, 0x3E, 0);
	
	si5341_write_register(SI5341_ADDRESS, 0x3F, 1);											// Assert PLL M divider update bit
	
	// Setup of the multisynth N-divider, and their numerators, for later use, and for a VCO frequency of 14 GHz
	set_page(3);
	si5341_write_register(SI5341_ADDRESS, 0x39, 0b00011111);								// Disable frequency updates by FINC/FDEC on all multisynth dividers 
	
	convert64to8 div_numerator;
	div_numerator.i64 = 14000000000;														// VCO frequency is 14.0 GHz

	si5341_write_consec(SI5341_ADDRESS, 0x02, 6, &div_numerator.byte[0]);					// Write numerator for multisynth divider 0
	si5341_write_consec(SI5341_ADDRESS, 0x0D, 6, &div_numerator.byte[0]);					// Write numerator for multisynth divider 1
	si5341_write_consec(SI5341_ADDRESS, 0x18, 6, &div_numerator.byte[0]);					// Write numerator for multisynth divider 2
	si5341_write_consec(SI5341_ADDRESS, 0x23, 6, &div_numerator.byte[0]);					// Write numerator for multisynth divider 3
}




/*
************************************************************************************************************************************************
Setting the Frequencies of LO1, LO2, and the BFO:
*/

static const uint32_t if1_freq = 86850000;													// IF 1 frequency 86.85 MHz (roofing filter center frequency)
static const uint32_t if2_freq = 10700000;													// IF 2 frequency 10.7 MHz
static const uint32_t if3_freq =   455000;													// IF 3 frequency 455 kHz for FM demodulation

static void lo1_set_freq(uint32_t freq)														// freq: Frequency in Hz
{
	convert32to8 div_denominator;
	div_denominator.i32 = freq * 2;
	
	set_page(3);
	si5341_write_consec(SI5341_ADDRESS, 0x08, 4, &div_denominator.byte[0]);					// Write denominator for multisynth divider 0
	si5341_write_register(SI5341_ADDRESS, 0x0C, 1);											// Assert update bit for multisynth divider 0
}

static void bfo_set_freq(uint32_t freq)														// freq: Frequency in Hz
{
	convert32to8 div_denominator;
	div_denominator.i32 = freq * 2;
	
	set_page(3);
	si5341_write_consec(SI5341_ADDRESS, 0x13, 4, &div_denominator.byte[0]);					// Write denominator for multisynth divider 1
	si5341_write_register(SI5341_ADDRESS, 0x17, 1);											// Assert update bit for multisynth divider 1
}

static void lo3_set_freq(uint32_t freq)														// freq: Frequency in Hz
{
	convert32to8 div_denominator;
	div_denominator.i32 = freq * 2;
	
	set_page(3);
	si5341_write_consec(SI5341_ADDRESS, 0x1E, 4, &div_denominator.byte[0]);					// Write denominator for multisynth divider 2
	si5341_write_register(SI5341_ADDRESS, 0x22, 1);											// Assert update bit for multisynth divider 2
}

static void lo2_set_freq(uint32_t freq)														// freq: Frequency in Hz
{
	convert32to8 div_denominator;
	div_denominator.i32 = freq * 2;
	
	set_page(3);
	si5341_write_consec(SI5341_ADDRESS, 0x29, 4, &div_denominator.byte[0]);					// Write denominator for multisynth divider 3
	si5341_write_register(SI5341_ADDRESS, 0x2D, 1);											// Assert update bit for multisynth divider 3
}

void freq_init(void)
{
	lo1_set_freq( 10000000 );																// Set LO1 initially to 10 MHz
	lo2_set_freq( if1_freq - if2_freq );													// Set LO2 to roofing filter center frequency minus IF 2 frequency
	lo3_set_freq( if2_freq - if3_freq );													// Set LO3 ti IF 2 frequency minus IF3 frequency
	bfo_set_freq( if2_freq );																// Set BFO to IF 2 frequency
}




/*
************************************************************************************************************************************************
Updating the Frequencies of LO1 and the BFO:
This function is called periodically in the main function.
*/

void freq_update(void)
{
	static uint32_t freq_old;
	static int16_t bfo_offs_old;

	if( tuned_freq != freq_old )															// tuned_freq has changed?
	{
		freq_old = tuned_freq;		
		lo1_set_freq( if1_freq - freq_old );												// Set LO1 to roofing filter center frequency minus tuned frequency
	}
	if( bfo_offset != bfo_offs_old )														// bfo_offset has changed?
	{
		bfo_offs_old = bfo_offset;		
		bfo_set_freq( if2_freq - bfo_offs_old );											// Set BFO to IF frequency minus BFO offset frequency
	}
}